##
## ppc.library emulation
## (c)1998-2000 Frank Wille <frank@phoenix.owl.de>
##
## Miscellaneous PPC assembler support functions.
##
## V0.6g (05.12.1999) phx
##       A stub routine for PPCprintf(), which does the required conversion
##       from V.4 ABI to PowerOpen ABI varargs.
## V0.6d (11.05.1999) phx
##       Flush PPC data cache. Also invalidate PPC instruction cache for
##       _ppc_wli function.
## V0.6b (18.04.1999) phx
##       Use '_' prefix for functions names. They must be callable from C.
## V0.4b (03.01.1999) phx
##       Moved ppccpuinfo into kernel_init().
## V0.4  (26.11.1998) phx
##       created.
##       ppc_[r/w][b/w/l] for reading/writing data by PPC.
##

	.file	"miscppc.s"
	.include "warpos_lvo.i"


	.extern	_kernPPCprintfPO



	.text


	.globl	_ppc_rb
_ppc_rb:
# r3 address
# returns: r3 data
	dcbf	0,r3
	sync
	lbz	r3,0(r3)
	blr


	.globl	_ppc_rw
_ppc_rw:
# r3 address
# returns: r3 data
	dcbf	0,r3
	sync
	lhz	r3,0(r3)
	blr

	.globl	_ppc_rl
_ppc_rl:
# r3 address
# returns: r3 data
	dcbf	0,r3
	sync
	lwz	r3,0(r3)
	blr


	.globl	_ppc_wb
_ppc_wb:
# r3 address
# r4 data
	stb	r4,0(r3)
	dcbst	0,r3
	sync
	blr


	.globl	_ppc_ww
_ppc_ww:
# r3 address
# r4 data
	sth	r4,0(r3)
	dcbst	0,r3
	sync
	blr


	.globl	_ppc_wl
_ppc_wl:
# r3 address
# r4 data
	stw	r4,0(r3)
	dcbst	0,r3
	sync
	blr


	.globl	_ppc_wli
_ppc_wli:
# r3 address
# r4 data
	stw	r4,0(r3)
	dcbst	0,r3
	sync
	icbi	0,r3
	sync
	isync
	blr


	.globl	_kernPPCprintf
_kernPPCprintf:
# The V.4 ABI varargs will be converted into a PowerOpen-style varargs
# list, which the kernPPCprintfPO() function in kernel.c is able to use.
# @@@ Float varargs are not supported, but PPCprintf() doesn't support
# them neither.
# r3 format string
# ... V.4 ABI varargs in r4-r10 and in stack frame
# r12 PPCLibBase
	mflr	r0
	stw	r0,8(r1)

# 64 bytes for storing PPCLibBase at offset 60,
# 56 bytes for saving 7 words, aligned to 16 bytes => 128
	stwu	r1,-128(r1)
	stw	r12,60(r1)	# for making __getbase() work

# copy 7 words from upper stack frame, which will be overwritten
# by our varargs
	addi	r11,r1,128+44-4
	addi	r12,r1,64-4
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)

# now copy 7 potential varargs in r4-r10 into the stack frame,
# directly before the varargs-area in the caller's stack
	addi	r11,r1,128+44-4
	stwu	r4,4(r11)
	stwu	r5,4(r11)
	stwu	r6,4(r11)
	stwu	r7,4(r11)
	stwu	r8,4(r11)
	stwu	r9,4(r11)
	stwu	r10,4(r11)

# call C function with a PowerOpen va_list in 2nd argument
	addi	r4,r1,128+44
	bl	_kernPPCprintfPO

# restore misused stack frame area and cleanup
	addi	r11,r1,64-4
	addi	r12,r1,128+44-4
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	lwzu	r0,4(r11)
	stwu	r0,4(r12)
	addi	r1,r1,128
	lwz	r0,8(r1)
	mtlr	r0
	blr
