/*
** ppc.library emulation
** (c)1998-2000 Frank Wille <frank@phoenix.owl.de>
**
** PPC attributes, task attributes, exception handlers, etc.
**
** V0.7d (22.02.2000) phx
**       Returns value of HID1 >> 28 for PPCINFOTAG_CPUPLL, as suggested
**       by Harry Sintonen. This seems to be what original PowerUp does!
** V0.7c (11.01.2000) phx
**       Ok, mea culpa. There *are* PPC boards with LC040 out, so I can't
**       use the FPU on the M68k side. Fixed PPCINFOTAG_CPUPLL again.
**       PPCGetAttrs() will always return a result, not only when
**       ti_Data is 0.
** V0.7b (07.01.2000) phx
**       PPCINFOTAG_CPUPLL should return a special pll code and not
**       the calculated pll directly (reported by Sebastian Jedruszkiewicz).
** V0.6e (14.05.1999) phx
**       PPCSetTaskAttrs(). Supported tags: PPCTASKINFOTAG_PRIORITY,
**        PPCTASKTAG_EXTUSERDATA, PPCTASKINFOTAG_INPUTHANDLE,
**        PPCTASKINFOTAG_OUTPUTHANDLE and PPCTASKINFOTAG_ERRORHANDLE.
** V0.6d (13.05.1999) phx
**       PPCGetAttrs() returns 0x57415250 ('WARP') for PPCINFOTAG_EMULATION
**       (=0x8001f0ff), which allows a program to distinguish the
**       emulation library from the original.
** V0.5d (04.04.1999) phx
**       PPCGetTaskAttrs(). Supported tags: PPCTASKINFOTAG_TASK, VALUEPTR,
**        NAME, PRIORITY, STACKSIZE, INPUTHANDLE, OUTPUTHANDLE, ERRORHANDLE,
**        ELFOBJECT, STARTUP_MSG, STARTUP_MSGDATA, STARTUPMSG_LENGTH,
**        STARTUP_MSGID, STARTUP_MSGPORT, SIGALLOC, SIGWAIT, SIGRECVD,
**        WAITFINISHTASK. The hook and PPC-registers are not supported.
** V0.4  (21.11.1998) phx
**       created, PPCGetAttr (without PPCINFOTAG_EXCEPTIONHOOK)
*/

#include "ppclibemu.h"
#include "errors.h"
#include "supp.h"
#include "taskobject.h"
#include "mpsema_protos.h"
#include <powerup/ppclib/ppc.h>
#include <powerup/ppclib/tasks.h>
#include <proto/utility.h>

#define PPCINFOTAG_EMULATION (TAG_USER + 0x1f0ff)



ULONG PPCGetTaskAttrs(__reg("a0")struct TaskObject *to,
                      __reg("a1")struct TagItem *ti,
                      __reg("a6")struct PPCLibBase *ppcbase)
{
  static const char *FN = "PPCGetTaskAttrs(): ";
  struct TagItem *tag;
  ULONG ret = 0;
  ULONG *vptr;

  Dprintf(ppcbase,"%sTaskObject=0x%08lx first tag=0x%08lx\n",
          FN,to,ti->ti_Tag);

  if (tag = __FindTagItem(PPCTASKINFOTAG_TASK,ti,ppcbase->UtilityBase))
    to = (struct TaskObject *)tag->ti_Data;
  obtainMPSemaphore68k(ppcbase,&to->mpSema);

  if (tag = __FindTagItem(PPCTASKINFOTAG_VALUEPTR,ti,ppcbase->UtilityBase))
    vptr = (ULONG *)tag->ti_Data;

  if (tag = __FindTagItem(PPCTASKINFOTAG_ALLTASK,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sALLTASK not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_HOOK,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sHOOK not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_NAME,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->n.ln_Name;

  if (tag = __FindTagItem(PPCTASKINFOTAG_PRIORITY,ti,ppcbase->UtilityBase))
    ret = (ULONG)((long)to->n.ln_Pri);

  if (tag = __FindTagItem(PPCTASKINFOTAG_STACKSIZE,ti,ppcbase->UtilityBase))
    ret = to->this->tp_StackSize;

  if (tag = __FindTagItem(PPCTASKINFOTAG_USEDSTACKSIZE,ti,ppcbase->UtilityBase))
    /*ret = (ULONG)to->initialstack - (ULONG)__get_r1();*/
    Dprintf(ppcbase,"%sUSEDSTACKSIZE not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_INPUTHANDLE,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->inputhandle;

  if (tag = __FindTagItem(PPCTASKINFOTAG_OUTPUTHANDLE,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->outputhandle;

  if (tag = __FindTagItem(PPCTASKINFOTAG_ERRORHANDLE,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->errorhandle;

  if (tag = __FindTagItem(PPCTASKINFOTAG_ELFOBJECT,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->elf;

  if (tag = __FindTagItem(PPCTASKINFOTAG_STARTUP_MSG,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->startup;

  if (tag = __FindTagItem(PPCTASKINFOTAG_STARTUP_MSGDATA,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->msg_data;

  if (tag = __FindTagItem(PPCTASKINFOTAG_STARTUP_MSGLENGTH,ti,ppcbase->UtilityBase))
    ret = to->msg_length;

  if (tag = __FindTagItem(PPCTASKINFOTAG_STARTUP_MSGID,ti,ppcbase->UtilityBase))
    ret = to->msg_id;

  if (tag = __FindTagItem(PPCTASKINFOTAG_MSGPORT,ti,ppcbase->UtilityBase))
    ret = (ULONG)to->ppc_port;

  if (tag = __FindTagItem(PPCTASKINFOTAG_SIGALLOC,ti,ppcbase->UtilityBase))
    ret = to->this->tp_Task.tc_SigAlloc;

  if (tag = __FindTagItem(PPCTASKINFOTAG_SIGWAIT,ti,ppcbase->UtilityBase))
    ret = to->this->tp_Task.tc_SigWait;

  if (tag = __FindTagItem(PPCTASKINFOTAG_SIGRECVD,ti,ppcbase->UtilityBase))
    ret = to->this->tp_Task.tc_SigRecvd;

  if (tag = __FindTagItem(PPCTASKINFOTAG_WAITFINISHTASK,ti,ppcbase->UtilityBase))
    if (!(to->flags & TskObjF_Async))
      ret = (ULONG)to->m68kParent;

  if (tag = __FindTagItem(PPCTASKINFOTAG_CACHEFLUSH,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sCACHEFLUSH not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_STATE,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sSTATE not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_EXCEPTIONHOOK,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sEXCEPTIONHOOK not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_PC,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sPC not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_MSR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sMSR not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_LR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sLR not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_XER,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sXER not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_CR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sCR not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_CTR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sCTR not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_FPSCR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sFPSCR not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_GPR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sGPR not supported\n",FN);

  if (tag = __FindTagItem(PPCTASKINFOTAG_FPR,ti,ppcbase->UtilityBase))
    Dprintf(ppcbase,"%sFPR not supported\n",FN);

  releaseMPSemaphore68k(ppcbase,&to->mpSema);
  Dprintf(ppcbase,"%sreturning 0x%08lx\n",FN,ret);
  return (ret);
}


ULONG PPCSetTaskAttrs(__reg("a0")struct TaskObject *to,
                      __reg("a1")struct TagItem *ti,
                      __reg("a6")struct PPCLibBase *ppcbase)
{
  extern BOOL ppctaskattr(struct PPCLibBase *,struct TaskObject *,
                          ULONG,ULONG,ULONG *,BOOL);
  static const char *FN = "PPCSetTaskAttrs(): ";
  struct TagItem *tstate=ti,*tag;
  ULONG ret = TRUE;
  ULONG *vptr;
  struct PPCArgs pa;


  Dprintf(ppcbase,"%sTaskObject=0x%08lx\n",FN,to);
  vptr = (ULONG *)__GetTagData(PPCTASKINFOTAG_VALUEPTR,NULL,ti,
                               ppcbase->UtilityBase);
  _memset(&pa,sizeof(struct PPCArgs),0);
  pa.PP_Code = (APTR)ppctaskattr;
  pa.PP_Flags = PPF_LINEAR;

  while (tag = __NextTagItem(&tstate,ppcbase->UtilityBase)) {
    switch (tag->ti_Tag) {
      case PPCTASKINFOTAG_VALUEPTR:
        break;
      default:
        Dprintf(ppcbase,"%stag=0x%08lx data=0x%08lx\n",
                FN,tag->ti_Tag,tag->ti_Data);
        pa.PP_Regs[0] = (ULONG)ppcbase;
        pa.PP_Regs[1] = (ULONG)to;
        pa.PP_Regs[2] = (ULONG)tag->ti_Tag;
        pa.PP_Regs[3] = tag->ti_Data;
        pa.PP_Regs[4] = (ULONG)vptr;
        pa.PP_Regs[5] = FALSE;
        __RunPPC(&pa,ppcbase->PowerPCBase);
        if (pa.PP_Regs[0] == FALSE)
          ret = FALSE;
        break;
    }
  }

  Dprintf(ppcbase,"%sret=%d\n",FN,(int)ret);
  return (ret);
}


ULONG PPCGetAttrs(__reg("a0")struct TagItem *ti,
                  __reg("a6")struct PPCLibBase *ppcbase)
{
  struct TagItem *tag;

  Dprintf(ppcbase,"PPCGetAttrs(): first tag = 0x%08lx\n",ti->ti_Data);

  if (tag = __FindTagItem(PPCINFOTAG_CPU,ti,ppcbase->UtilityBase))
    return (ppcbase->ppc_PVR >> 16);

  if (tag = __FindTagItem(PPCINFOTAG_CPUREV,ti,ppcbase->UtilityBase))
    return (ppcbase->ppc_PVR & 0xffff);

  if (tag = __FindTagItem(PPCINFOTAG_CPUCOUNT,ti,ppcbase->UtilityBase))
    return (1);  /* 1 CPU is hard-coded */

  if (tag = __FindTagItem(PPCINFOTAG_CPUCLOCK,ti,ppcbase->UtilityBase))
    return ((ppcbase->ppc_CPUClock+500000)/1000000);

  if (tag = __FindTagItem(PPCINFOTAG_CPUPLL,ti,ppcbase->UtilityBase))
    return (ppcbase->ppc_HID1 >> 28);

  if (tag = __FindTagItem(PPCINFOTAG_EXCEPTIONHOOK,ti,ppcbase->UtilityBase)) {
    Dprintf(ppcbase,"PPCGetAttrs(): EXCEPTIONHOOK not supported\n");
    return (0);
  }

  if (tag = __FindTagItem(PPCINFOTAG_EMULATION,ti,ppcbase->UtilityBase))
    return (0x57415250);  /* return 'WARP', if in emulation mode */

  return (0);
}
